using Cairo;
using Gdk;
using Gtk;
using static System.Math;

using Rectangle = Cairo.Rectangle;

delegate void Notify();

// A model class for a game of Tic Tac Toe.
class Board {
    int[,] square = new int[3, 3];    // 0 = empty, 1 = player 1, 2 = player 2
    int moves = 0;  // number of moves on board

    public int player = 1;     // whose turn it is
    public int won = 0;     // who has won
    public int win_x, win_y, win_dx, win_dy;    // vector indicating winning squares
    
    public event Notify? changed;    // fires whenever the board changes
    
    public int this[int x, int y] {
        get => square[x, y];
    }

    public int? winner() {
        if (won != 0)
            return won;
        if (moves == 9)
            return 0;
        return null;
    }

    // Return true if the current player has three in a row,
    // starting at (x, y) and moving in direction (dx, dy).
    bool all(int x, int y, int dx, int dy) {
        for (int i = 0 ; i < 3 ; ++i) 
            if (square[x + i * dx, y + i * dy] != player)
                return false;
        
        win_x = x; win_y = y; win_dx = dx; win_dy = dy;
        return true;
    }

    // Return true if the current player has won.
    bool checkWin() {
        // check rows and columns
        for (int i = 0 ; i < 3 ; ++i)
            if (all(i, 0, 0, 1) || all(0, i, 1, 0))
                return true;
                
        // check diagonals
        return all(0, 0, 1, 1) || all(2, 0, -1, 1);
    }
    
    // Make the current player play at (x, y).    Return true if the move was legal.
    public bool move((int x, int y) m) {
        if (square[m.x, m.y] > 0)    // square is occupied
            return false;
            
        square[m.x, m.y] = player;
        moves += 1;
        if (checkWin())
            won = player;
        else player = 3 - player;
        
        if (changed != null)
            changed();
        return true;
    }
}

// A graphical interface for Tic Tac Toe.

class View : DrawingArea {
    const int SQUARE_SIZE = 100;
    const int MARGIN = SQUARE_SIZE / 2;
    
    Board board;
    
    void init() {
        QueueDraw();     // draw the initial position
        board.changed += QueueDraw;
    }
    
    public View() {
        board = new Board();
        init();
        AddEvents((int) EventMask.ButtonPressMask);    // ask to receive button press events
    }

    void drawLine(Context c, double x1, double y1, double x2, double y2) {
        c.MoveTo(x1, y1);
        c.LineTo(x2, y2);
        c.Stroke();
    }

    void drawX(Context c, Rectangle r) {
        drawLine(c, r.X, r.Y, r.X + r.Width, r.Y + r.Height);
        drawLine(c, r.X + r.Width, r.Y, r.X, r.Y + r.Height);
    }
    
    void drawO(Context c, Rectangle r) {
        c.Arc(r.X + r.Width / 2, r.Y + r.Height / 2, r.Width / 2, 0.0, 2 * PI);
        c.Stroke();
    }
    
    protected override bool OnButtonPressEvent(EventButton ev) {
        if (ev.Type == EventType.ButtonPress)
            if (board.winner() != null) {
                board = new Board();
                init();     // start a new game
            }
            else {
                int x = ((int) (ev.X - MARGIN)) / SQUARE_SIZE;
                int y = ((int) (ev.Y - MARGIN)) / SQUARE_SIZE;
                board.move((x, y));
            }

        return true;
    }

    protected override bool OnDrawn(Context c) {
        // Set a transformation matrix so that squares are 1 unit tall/wide, with a
        // MARGIN around the drawing area.
        c.Translate(MARGIN, MARGIN);
        c.Scale(SQUARE_SIZE, SQUARE_SIZE);
        
        // Adjust the line width to compensate for the scaling tranformation.
        c.LineWidth /= SQUARE_SIZE;
        
        if (board.won > 0) {
            c.SetSourceRGB(0.0, 1.0, 0.0);  // green
            for (int i = 0 ; i < 3 ; ++i) {
                c.Rectangle(board.win_x + i * board.win_dx,
                                        board.win_y + i * board.win_dy,
                                        1, 1);
                c.Fill();
            }
            c.SetSourceRGB(0.0, 0.0, 0.0);  // black
        }
        
        // Draw the tic tac toe grid.
        for (int i = 1 ; i <= 2 ; ++i) {
            drawLine(c, i, 0, i, 3);
            drawLine(c, 0, i, 3, i);
        }
        
        // Draw the X's and O's.
        for (int x = 0 ; x < 3 ; ++x)
            for (int y = 0 ; y < 3 ; ++y) {
                Rectangle rect = new Rectangle(x + 0.1, y + 0.1, 0.8, 0.8);
                
                switch (board[x, y]) {
                    case 1: drawX(c, rect); break;
                    case 2: drawO(c, rect); break;
                }
            }
        return true;
    }
}

class Frame : Gtk.Window {
    Frame() : base("tic tac toe") {
        Resize(400, 400);
        Add(new View());
    }
    
    protected override bool OnDeleteEvent(Event e) {
        Application.Quit();
        return true;
    }
    
    static void Main() {
        Application.Init();
        new Frame().ShowAll();
        Application.Run();
    }
}
